@@ -2,10 +2,10 @@ package CGI::Lingua;
 
 use warnings;
 use strict;
-use Carp;
+use Class::Autouse qw{Carp Locale::Language Locale::Object::Country Locale::Object::DB I18N::AcceptLanguage I18N::LangTags::Detect};
 
 use vars qw($VERSION);
-our $VERSION = '0.51';
+our $VERSION = '0.52';
 
 =head1 NAME
 
@@ -13,7 +13,7 @@ CGI::Lingua - Create a multilingual web page
 
 =head1 VERSION
 
-Version 0.51
+Version 0.52
 
 =cut
 
@@ -108,7 +108,7 @@ sub new {
 		# croak('You must give a list of supported languages');
 	# }
 	unless($params{supported}) {
-		croak('You must give a list of supported languages');
+		Carp::croak('You must give a list of supported languages');
 	}
 
 	return bless {
@@ -126,6 +126,7 @@ sub new {
 		_logger => $params{logger},
 		_have_ipcountry => -1,	# -1 = don't know
 		_have_geoip => -1,	# -1 = don't know
+		_have_geoipfree => -1,	# -1 = don't know
 	}, $class;
 }
 
@@ -150,7 +151,7 @@ sub _warn {
 	if($self->{_logger}) {
 		$self->{_logger}->warn($warning);
 	} elsif(!defined($self->{_syslog})) {
-		carp($warning);
+		Carp::carp($warning);
 	}
 }
 
@@ -216,7 +217,7 @@ on a site that only serves British English, sublanguage() will return undef.
 sub sublanguage {
 	my $self = shift;
 
-	unless($self->{_sublanguage}) {
+	unless($self->{_slanguage}) {
 		$self->_find_language();
 	}
 	return $self->{_sublanguage};
@@ -235,7 +236,7 @@ language_code_alpha2() returns undef.
 sub language_code_alpha2 {
 	my $self = shift;
 
-	unless($self->{_slanguage_code_alpha2}) {
+	unless($self->{_slanguage}) {
 		$self->_find_language();
 	}
 	return $self->{_slanguage_code_alpha2};
@@ -264,7 +265,7 @@ when you've asked for en-gb, or undef.
 sub sublanguage_code_alpha2 {
 	my $self = shift;
 
-	unless($self->{_sublanguage_code_alpha2}) {
+	unless($self->{_slanguage}) {
 		$self->_find_language();
 	}
 	return $self->{_sublanguage_code_alpha2};
@@ -287,22 +288,20 @@ sub requested_language {
 	return $self->{_rlanguage};
 }
 
+# The language cache is stored as country_2_letter -> $language_human_readable_name=$language_2_letter
+# The IP cache is stored as ip -> country_human_readable_name
 sub _find_language {
 	my $self = shift;
 
 	$self->{_rlanguage} = 'Unknown';
 	$self->{_slanguage} = 'Unknown';
 
-	require Locale::Object::Country;
-	Locale::Object::Country->import();
-
 	# Use what the client has said
-	if($ENV{'HTTP_ACCEPT_LANGUAGE'}) {
-		require I18N::AcceptLanguage;
-		require Locale::Language;
-
-		I18N::AcceptLanguage->import();
-		Locale::Language->import();
+	my $http_accept_language = $ENV{'HTTP_ACCEPT_LANGUAGE'};
+	if($http_accept_language) {
+		if($self->{_logger}) {
+			$self->{_logger}->debug("HTTP_ACCEPT_LANGUAGE: $http_accept_language");
+		}
 
 		# Workaround for RT 74338
 		local $SIG{__WARN__} = sub {
@@ -310,128 +309,189 @@ sub _find_language {
 				warn $_[0];
 			}
 		};
-		my $l = I18N::AcceptLanguage->new()->accepts($ENV{'HTTP_ACCEPT_LANGUAGE'}, $self->{_supported});
-		$SIG{__WARN__} = 'DEFAULT';
-		if((!$l) && ($ENV{'HTTP_ACCEPT_LANGUAGE'} =~ /(.+)-.+/)) {
+		my $i18n = I18N::AcceptLanguage->new();
+		my $l = $i18n->accepts($http_accept_language, $self->{_supported});
+		local $SIG{__WARN__} = 'DEFAULT';
+		if((!$l) && ($http_accept_language =~ /(.+)-.+/)) {
 			# Fall back position, e,g. we want US English on a site
 			# only giving British English, so allow it as English.
 			# The calling program can detect that it's not the
 			# wanted flavour of English by looking at
 			# requested_language
-			if(I18N::AcceptLanguage->new()->accepts($1, $self->{_supported})) {
+			if($i18n->accepts($1, $self->{_supported})) {
 				$l = $1;
 			}
 		}
 		if($l) {
-			$self->{_slanguage} = Locale::Language::code2language($l);
-			if($self->{_slanguage}) {
-				# We have the language, but not the right
-				# sublanguage, e.g. they want US English but we
-				# only support British English or English
-				# wanted: en-us, got en-gb and en
-				$self->{_slanguage_code_alpha2} = $l;
-				$self->{_rlanguage} = $self->{_slanguage};
+			if($self->{_logger}) {
+				$self->{_logger}->debug("l: $l");
+			}
 
-				if($ENV{'HTTP_ACCEPT_LANGUAGE'} =~ /..-(..)$/) {
-					my $l = Locale::Object::Country->new(code_alpha2 => $1);
+			unless($l =~ /^..-..$/) {
+				$self->{_slanguage} = $self->_code2language($l);
+				if($self->{_slanguage}) {
+					if($self->{_logger}) {
+						$self->{_logger}->debug("_slanguage: $self->{_slanguage}");
+					}
+					# We have the language, but not the right
+					# sublanguage, e.g. they want US English but we
+					# only support British English or English
+					# wanted: en-us, got en-gb and en
+					$self->{_slanguage_code_alpha2} = $l;
+					$self->{_rlanguage} = $self->{_slanguage};
+
+					my $l;
+					if($http_accept_language =~ /..-(..)$/) {
+						$l = $self->_code2country($1);
+					} elsif($http_accept_language =~ /..-([a-z]{2,3})$/i) {
+						$l = Locale::Object::Country->new(code_alpha3 => $1);
+					}
 					if($l) {
 						$self->{_rlanguage} .= ' (' . $l->name . ')';
 						# The requested sublanguage
 						# isn't supported so don't
 						# define that
 					}
-				} elsif($ENV{'HTTP_ACCEPT_LANGUAGE'} =~ /..-([a-z]{2,3})$/i) {
-					my $l = Locale::Object::Country->new(code_alpha3 => $1);
-					if($l) {
-						$self->{_rlanguage} .= ' (' . $l->name . ')';
-						# The requested sublanguage isn't
-						# supported so don't define that
-					}
+					return;
 				}
-				return;
 			}
 			# TODO: Handle es-419 "Spanish (Latin America)"
 			if($l =~ /(.+)-(..)$/) {
 				my $alpha2 = $1;
 				my $variety = $2;
-				my $accepts = I18N::AcceptLanguage->new()->accepts($l, $self->{_supported});
+				# my $accepts = $i18n->accepts($l, $self->{_supported});
+				my $accepts = $l;
 
 				if($accepts) {
-					$self->{_slanguage} = Locale::Language::code2language($accepts);
-					if(length($variety) == 2) {
-						my $c = Locale::Object::Country->new(code_alpha2 => $variety);
-						if(defined($c)) {
-							$self->{_sublanguage} = $c->name;
+					if($self->{_logger}) {
+						$self->{_logger}->debug("accepts: $accepts");
+					}
+					if($accepts =~ /\-/) {
+						delete $self->{_slanguage};
+					} else  {
+						my $from_cache;
+						if($self->{_cache}) {
+							$from_cache = $self->{_cache}->get($accepts);
+						}
+						if($from_cache) {
+							if($self->{_logger}) {
+								$self->{_logger}->debug("$accepts is in cache as $from_cache");
+							}
+							$self->{_rlanguage} = (split(/=/, $from_cache))[0];
+						} else {
+							$self->{_slanguage} = $self->_code2language($accepts);
+						}
+						if($self->{_slanguage}) {
+							if($variety eq 'uk') {
+								# ???
+								$self->_warn({
+									warning => "Resetting country code to GB for $http_accept_language"
+								});
+								$variety = 'gb';
+							}
+							my $c = $self->_code2countryname($variety);
+							if(defined($c)) {
+								$self->{_sublanguage} = $c;
+							}
+							$self->{_slanguage_code_alpha2} = $accepts;
+							if($self->{_sublanguage}) {
+								$self->{_rlanguage} = "$self->{_slanguage} ($self->{_sublanguage})";
+								if($self->{_logger}) {
+									$self->{_logger}->debug("rlanguage: $self->{_rlanguage}");
+								}
+							}
+							$self->{_sublanguage_code_alpha2} = $variety;
+							unless($from_cache) {
+								if($self->{_logger}) {
+									$self->{_logger}->debug("Set $variety to $self->{_slanguage}=$self->{_slanguage_code_alpha2}");
+								}
+								$self->{_cache}->set($variety, "$self->{_slanguage}=$self->{_slanguage_code_alpha2}", '1 month');
+							}
+							return;
 						}
 					}
-					if($self->{_slanguage}) {
-						$self->{_slanguage_code_alpha2} = $accepts;
-						if($self->{_sublanguage}) {
-							$self->{_rlanguage} = "$self->{_slanguage} ($self->{_sublanguage})";
-						}
-						$self->{_sublanguage_code_alpha2} = $variety;
-						return;
-					}
 				}
-				my $lang = Locale::Language::code2language($alpha2);
-				unless($lang) {
-					$lang = $1;
+				$self->{_rlanguage} = $self->_code2language($alpha2);
+				if($self->{_logger}) {
+					$self->{_logger}->debug("_rlanguage: $self->{_rlanguage}");
 				}
-				$self->{_rlanguage} = $lang;
-				$self->_get_closest($alpha2, $alpha2);
-				if($self->{_sublanguage}) {
-					$ENV{'HTTP_ACCEPT_LANGUAGE'} =~ /(.{2})-(..)/;
+				if($accepts) {
+					$http_accept_language =~ /(.{2})-(..)/;
 					$variety = lc($2);
 					# Ignore en-029 etc (Carribean English)
-					if($variety =~ /[a-z]{2,3}/) {
+					if(($variety =~ /[a-z]{2,3}/) && !defined($self->{_sublanguage})) {
+						$self->_get_closest($alpha2, $alpha2);
+						if($self->{_logger}) {
+							$self->{_logger}->debug("Find the country code for $variety");
+						}
 						if($variety eq 'uk') {
 							# ???
 							$self->_warn({
-								warning => "Resetting country code to GB for $ENV{'HTTP_ACCEPT_LANGUAGE'}"
+								warning => "Resetting country code to GB for $http_accept_language"
 							});
 							$variety = 'gb';
 						}
-						my $db = Locale::Object::DB->new();
-						my @results = $db->lookup(
-							table => 'country',
-							result_column => '*',
-							search_column => 'code_alpha2',
-							value => $variety
-						);
-						if(defined($results[0])) {
-							eval {
-								$lang = Locale::Object::Country->new(code_alpha2 => $variety);
-							};
-						} else {
-							$lang = undef;
+						my $from_cache;
+						my $language_name;
+						if($self->{_cache}) {
+							$from_cache = $self->{_cache}->get($variety);
 						}
-						if($@ || !defined($lang)) {
+						if(defined($from_cache)) {
+							if($self->{_logger}) {
+								$self->{_logger}->debug("$variety is in cache as $from_cache");
+							}
+							my $language_code2;
+							($language_name, $language_code2) = split(/=/, $from_cache);
+							$language_name = $self->_code2countryname($variety);
+						} else {
+							my $db = Locale::Object::DB->new();
+							my @results = @{$db->lookup(
+								table => 'country',
+								result_column => 'name',
+								search_column => 'code_alpha2',
+								value => $variety
+							)};
+							if(defined($results[0])) {
+								eval {
+									$language_name = $self->_code2countryname($variety);
+								}
+							}
+						}
+						if($@ || !defined($language_name)) {
 							$self->{_sublanguage} = 'Unknown';
 							$self->_warn({
-								warning => "Can't determine values for $ENV{'HTTP_ACCEPT_LANGUAGE'}"
+								warning => "Can't determine values for $http_accept_language"
 							});
 						} else {
-							$self->{_sublanguage} = $lang->name;
-						}
-						if(defined($self->{_sublanguage})) {
-							$self->{_rlanguage} = "$self->{_slanguage} ($self->{_sublanguage})";
-							$self->{_sublanguage_code_alpha2} = $variety;
-							return;
+							$self->{_sublanguage} = $language_name;
+							if($self->{_logger}) {
+								$self->{_logger}->debug('variety name ' . $self->{_sublanguage});
+							}
+							if($self->{_cache} && !defined($from_cache)) {
+								if($self->{_logger}) {
+									$self->{_logger}->debug("Set $variety to $self->{_slanguage}=$self->{_slanguage_code_alpha2}");
+								}
+								$self->{_cache}->set($variety, "$self->{_slanguage}=$self->{_slanguage_code_alpha2}", '1 month');
+							}
 						}
 					}
+					if(defined($self->{_sublanguage})) {
+						$self->{_rlanguage} = "$self->{_slanguage} ($self->{_sublanguage})";
+						$self->{_sublanguage_code_alpha2} = $variety;
+						return;
+					}
 				}
 			}
 		}
 		if($self->{_slanguage} && ($self->{_slanguage} ne 'Unknown')) {
 			if($self->{_rlanguage} eq 'Unknown') {
-				require I18N::LangTags::Detect;
 				$self->{_rlanguage} = I18N::LangTags::Detect::detect();
 			}
 			if($self->{_rlanguage}) {
-				my $l = Locale::Language::code2language($self->{_rlanguage});
+				my $l = $self->_code2language($self->{_rlanguage});
 				if($l) {
 					$self->{_rlanguage} = $l;
-				} else {
+				# } else {
 					# We have the language, but not the right
 					# sublanguage, e.g. they want US English but we
 					# only support British English
@@ -455,46 +515,101 @@ sub _find_language {
 	# address for an alternative
 	my $country = $self->country();
 	if(defined($country)) {
+		if($self->{_logger}) {
+			$self->{_logger}->debug("country: $country");
+		}
 		# Determine the first official language of the country
 
-		my $l = Locale::Object::Country->new(code_alpha2 => uc($country));
-		if($l) {
-			$l = ($l->languages_official)[0];
+		my $from_cache;
+		if($self->{_cache}) {
+			$from_cache = $self->{_cache}->get($country);
+		}
+		my $language_name;
+		my $language_code2;
+		if($from_cache) {
+			if($self->{_logger}) {
+				$self->{_logger}->debug("$country is in cache as $from_cache");
+			}
+			($language_name, $language_code2) = split(/=/, $from_cache);
+		} else {
+			my $l = $self->_code2country(uc($country));
+			if($l) {
+				$l = ($l->languages_official)[0];
+				if(defined($l)) {
+					$language_name = $l->name;
+					$language_code2 = $l->code_alpha2;
+					if($self->{_logger} && $language_name) {
+						$self->{_logger}->debug("Official language: $language_name");
+					}
+				}
+			}
 		}
 		my $ip = $ENV{'REMOTE_ADDR'};
-		if($l && $l->name) {
-			$self->{_rlanguage} = $l->name;
+		if($language_name) {
+			$self->{_rlanguage} = $language_name;
 			unless((exists($self->{_slanguage})) && ($self->{_slanguage} ne 'Unknown')) {
 				# Check if the language is one that we support
 				# Don't bother with secondary language
-				require Locale::Language;
+				my $code;
 
-				Locale::Language->import();
-
-				my $code = Locale::Language::language2code($self->{_rlanguage});
-				unless($code) {
-					if($ENV{'HTTP_ACCEPT_LANGUAGE'}) {
-						$code = Locale::Language::language2code($ENV{'HTTP_ACCEPT_LANGUAGE'});
+				if($language_name && $language_code2 && !defined($http_accept_language)) {
+					# This sort of thing speeds up search engine access a lot
+					if($self->{_logger}) {
+						$self->{_logger}->debug("Fast assign to $language_code2");
 					}
+					$code = $language_code2;
+				} else {
+					if($self->{_logger}) {
+						$self->{_logger}->debug("Call language2code on $self->{_rlanguage}");
+					}
+					$code = Locale::Language::language2code($self->{_rlanguage});
 					unless($code) {
-						# If language is Norwegian (Nynorsk)
-						# lookup Norwegian
-						if($self->{_rlanguage} =~ /(.+)\s\(.+/) {
-							$code = Locale::Language::language2code($1);
+						if($http_accept_language && ($http_accept_language ne $self->{_rlanguage})) {
+							if($self->{_logger}) {
+								$self->{_logger}->debug("Call language2code on $http_accept_language");
+							}
+							$code = Locale::Language::language2code($http_accept_language);
 						}
 						unless($code) {
-							$self->_warn({
-								warning => "Can\'t determine code from IP $ip for requested language $self->{_rlanguage}"
-							});
+							# If language is Norwegian (Nynorsk)
+							# lookup Norwegian
+							if($self->{_rlanguage} =~ /(.+)\s\(.+/) {
+								if((!defined($http_accept_language)) || ($1 ne $self->{_rlanguage})) {
+									if($self->{_logger}) {
+										$self->{_logger}->debug("Call language2code on $1");
+									}
+									$code = Locale::Language::language2code($1);
+								}
+							}
+							unless($code) {
+								$self->_warn({
+									warning => "Can't determine code from IP $ip for requested language $self->{_rlanguage}"
+								});
+							}
 						}
 					}
 				}
 				if($code) {
-					$self->_get_closest($code, $l->code_alpha2);
+					$self->_get_closest($code, $language_code2);
+					unless($self->{_slanguage}) {
+						$self->_warn({
+							warning => "Couldn't determine closest language for $language_name in $self->{_supported}"
+						});
+					} elsif($self->{_logger}) {
+						$self->{_logger}->debug("language set to $self->{_slanguage}, code set to $code");
+					}
 				}
 			}
-			if($self->{_cache} && defined($ip)) {
-				$country = $self->{_cache}->set($ip, $country, '1 hour');
+			if(!defined($self->{_slanguage_code_alpha2})) {
+				if($self->{_logger}) {
+					$self->{_logger}->debug("Can't determine slanguage_code_alpha2");
+				};
+			} elsif(!defined($from_cache) && $self->{_cache} &&
+			   defined($self->{_slanguage_code_alpha2})) {
+				if($self->{_logger}) {
+					$self->{_logger}->debug("Set $country to $language_name=$self->{_slanguage_code_alpha2}");
+				}
+				$self->{_cache}->set($country, "$language_name=$self->{_slanguage_code_alpha2}", '1 month');
 			}
 		} elsif(defined($ip)) {
 			$self->_warn({
@@ -502,6 +617,7 @@ sub _find_language {
 			});
 		}
 	}
+	return;
 }
 
 # Try our very best to give the right country - if they ask for en-us and
@@ -523,15 +639,16 @@ sub _get_closest {
 			last;
 		}
 	}
+	return;
 }
 
 =head2 country
 
 Returns the two character country code of the remote end in lower case.
 
-If L<IP::Country> or L<Geo::IP> is installed, CGI::Lingua will make
-use of that, otherwise it will do a Whois lookup.
-If you do not have either installed I recommend you make use of the
+If L<IP::Country>, L<Geo::IPfree> or L<Geo::IP> is installed,
+CGI::Lingua will make use of that, otherwise it will do a Whois lookup.
+If you do not have any of those installed I recommend you make use of the
 caching capability of CGI::Lingua.
 
 =cut
@@ -551,12 +668,31 @@ sub country {
 		$self->{_country} = lc($ENV{'GEOIP_COUNTRY_CODE'});
 		return $self->{_country};
 	}
+	if(($ENV{'HTTP_CF_IPCOUNTRY'}) && ($ENV{'HTTP_CF_IPCOUNTRY'} ne 'XX')) {
+		# Hosted by Cloudfare
+		$self->{_country} = lc($ENV{'HTTP_CF_IPCOUNTRY'});
+		return $self->{_country};
+	}
 
 	my $ip = $ENV{'REMOTE_ADDR'};
 
 	unless($ip) {
 		return();
 	}
+	if($self->{_cache}) {
+		$self->{_country} = $self->{_cache}->get($ip);
+		if($self->{_logger}) {
+			if(defined($self->{_country})) {
+				$self->{_logger}->trace("Get $ip from cache = $self->{_country}");
+			} else {
+				$self->{_logger}->trace("$ip isn't in the cache");
+			}
+		}
+		if(defined($self->{_country})) {
+			return $self->{_country};
+		}
+	}
+
 	require Data::Validate::IP;
 	Data::Validate::IP->import();
 
@@ -572,111 +708,104 @@ sub country {
 		}
 	}
 
-	if($self->{_cache}) {
-		$self->{_country} = $self->{_cache}->get($ip);
-		if($self->{_logger}) {
-			if(defined($self->{_country})) {
-				$self->{_logger}->trace("Get $ip from cache = $self->{_country}");
+	if($self->{_have_ipcountry} == -1) {
+		if(eval { require IP::Country; }) {
+			IP::Country->import();
+			$self->{_have_ipcountry} = 1;
+			$self->{_ipcountry} = IP::Country::Fast->new();
+		} else {
+			$self->{_have_ipcountry} = 0;
+		}
+	}
+	if($self->{_logger}) {
+		$self->{_logger}->debug("have_ipcountry $self->{_have_ipcountry}");
+	}
+
+	if($self->{_have_ipcountry}) {
+		$self->{_country} = $self->{_ipcountry}->inet_atocc($ip);
+		if($self->{_country}) {
+			$self->{_country} = lc($self->{_country});
+		} else {
+			$self->_warn({
+				warning => "$ip is not known by IP::Country"
+			});
+		}
+	}
+	unless(defined($self->{_country})) {
+		if($self->{_have_geoip} == -1) {
+			if(eval { require Geo::IP; }) {
+				Geo::IP->import();
+				$self->{_have_geoip} = 1;
+				# GEOIP_STANDARD = 0, can't use that because you'll
+				# get a syntax error
+				$self->{_geoip} = Geo::IP->new(0);
 			} else {
-				$self->{_logger}->trace("$ip isn't in the cache");
+				$self->{_have_geoip} = 0;
+			}
+		}
+		if($self->{_have_geoip} == 1) {
+			$self->{_country} = $self->{_geoip}->country_code_by_addr($ip);
+		}
+		unless(defined($self->{_country})) {
+			if($self->{_have_geoipfree} == -1) {
+				if(eval { require Geo::IPfree; }) {
+					Geo::IPfree::IP->import();
+					$self->{_have_geoipfree} = 1;
+					$self->{_geoipfree} = Geo::IPfree->new();
+				} else {
+					$self->{_have_geoipfree} = 0;
+				}
+			}
+			if($self->{_have_geoipfree} == 1) {
+				$self->{_country} = lc(($self->{_geoipfree}->LookUp($ip))[0]);
 			}
 		}
 	}
+	if($self->{_country} && ($self->{_country} eq 'eu')) {
+		delete($self->{_country});
+	}
+	unless($self->{_country}) {
+		require Net::Whois::IP;
+		Net::Whois::IP->import();
 
-	unless(defined($self->{_country})) {
-		if(($ENV{'HTTP_CF_IPCOUNTRY'}) && ($ENV{'HTTP_CF_IPCOUNTRY'} ne 'XX')) {
-			# Hosted by Cloudfare
-			$self->{_country} = lc($ENV{'HTTP_CF_IPCOUNTRY'});
-		} else {
-			if($self->{_have_ipcountry} == -1) {
-				if(eval { require IP::Country; }) {
-					IP::Country->import();
-					$self->{_have_ipcountry} = 1;
-					$self->{_ipcountry} = IP::Country::Fast->new();
-				} else {
-					$self->{_have_ipcountry} = 0;
-				}
-			}
-			if($self->{_logger}) {
-				$self->{_logger}->debug("have_ipcountry $self->{_have_ipcountry}");
-			}
+		my $whois;
 
-			if($self->{_have_ipcountry} == 1) {
-				$self->{_country} = $self->{_ipcountry}->inet_atocc($ip);
-				if($self->{_country}) {
-					$self->{_country} = lc($self->{_country});
-				} else {
-					$self->_warn({
-						warning => "$ip is not known by IP::Country"
-					});
-				}
+		eval {
+			# Catch connection timeouts to
+			# whois.ripe.net by turning the carp
+			# into an error
+			local $SIG{__WARN__} = sub { die $_[0] };
+			$whois = Net::Whois::IP::whoisip_query($ip);
+		};
+		unless($@ || !defined($whois) || (ref($whois) ne 'HASH')) {
+			if(defined($whois->{Country})) {
+				$self->{_country} = $whois->{Country};
+			} elsif(defined($whois->{country})) {
+				$self->{_country} = $whois->{country};
 			}
-			unless(defined($self->{_country})) {
-				if($self->{_have_geoip} == -1) {
-					if(eval { require Geo::IP; }) {
-						Geo::IP->import();
-						$self->{_have_geoip} = 1;
-						# GEOIP_STANDARD = 0, can't use that because you'll
-						# get a syntax error
-						$self->{_geoip} = Geo::IP->new(0);
-					} else {
-						$self->{_have_geoip} = 0;
-					}
-				}
-				if($self->{_have_geoip} == 1) {
-					$self->{_country} = $self->{_geoip}->country_code_by_addr($ip);
-				}
-			}
-		}
-		if($self->{_country} && ($self->{_country} eq 'eu')) {
-			delete($self->{_country});
-		}
-		unless($self->{_country}) {
-			require Net::Whois::IP;
-			Net::Whois::IP->import();
-
-			my $whois;
-
-			eval {
-				# Catch connection timeouts to
-				# whois.ripe.net by turning the carp
-				# into an error
-				local $SIG{__WARN__} = sub { die $_[0] };
-				$whois = Net::Whois::IP::whoisip_query($ip);
-			};
-			unless($@ || !defined($whois) || (ref($whois) ne 'HASH')) {
-				if(defined($whois->{Country})) {
-					$self->{_country} = $whois->{Country};
-				} elsif(defined($whois->{country})) {
-					$self->{_country} = $whois->{country};
-				}
-				if($self->{_country} && ($self->{_country} eq 'eu')) {
-					delete($self->{_country});
-				}
-			}
-
-			unless($self->{_country}) {
-				require Net::Whois::IANA;
-				Net::Whois::IANA->import();
-
-				my $iana = new Net::Whois::IANA;
-				eval {
-					$iana->whois_query(-ip => $ip);
-				};
-				unless ($@) {
-					$self->{_country} = $iana->country();
-				}
-			}
-			if($self->{_country} eq 'eu') {
+			if($self->{_country} && ($self->{_country} eq 'eu')) {
 				delete($self->{_country});
 			}
-			if($self->{_country}) {
-				# 190.24.1.122 has carriage return in its WHOIS record
-				$self->{_country} =~ s/[\r\n]//g;
-				if($self->{_country} =~ /^(..)\s*#.*/) {
-					# Remove comments in the Whois record
-					$self->{_country} = $1;
-				}
+		}
+
+		unless($self->{_country}) {
+			require Net::Whois::IANA;
+			Net::Whois::IANA->import();
+
+			my $iana = new Net::Whois::IANA;
+			eval {
+				$iana->whois_query(-ip => $ip);
+			};
+			unless ($@) {
+				$self->{_country} = $iana->country();
+			}
+		}
+		if($self->{_country}) {
+			# 190.24.1.122 has carriage return in its WHOIS record
+			$self->{_country} =~ s/[\r\n]//g;
+			if($self->{_country} =~ /^(..)\s*#/) {
+				# Remove comments in the Whois record
+				$self->{_country} = $1;
 			}
 		}
 	}
@@ -699,14 +828,14 @@ sub country {
 				$self->_warn({
 					warning => "$ip has country of eu"
 				});
-				delete($self->{_country});
+				$self->{_country} = 'Unknown';
 			}
 		}
 		if($self->{_cache}) {
-			$self->{_cache}->set($ip, $self->{_country}, '1 hour');
 			if($self->{_logger}) {
-				$self->{_logger}->debuf("Set $ip to $self->{_country}");
+				$self->{_logger}->debug("Set $ip to $self->{_country}");
 			}
+			$self->{_cache}->set($ip, $self->{_country}, '1 hour');
 		}
 	}
 
@@ -735,8 +864,6 @@ sub locale {
 	if($self->{_locale}) {
 		return $self->{_locale};
 	}
-	require Locale::Object::Country;
-	Locale::Object::Country->import();
 
 	# First try from the User Agent.  Probably only works with Mozilla and
 	# Safari.  I don't know about Opera.  It won't work with IE or Chrome.
@@ -750,7 +877,7 @@ sub locale {
 			$candidate =~ s/\s$//g;
 			if($candidate =~ /^[a-zA-Z]{2}-([a-zA-Z]{2})$/) {
 				local $SIG{__WARN__} = undef;
-				my $c = Locale::Object::Country->new(code_alpha2 => $1);
+				my $c = $self->_code2country($1);
 				if($c) {
 					$self->{_locale} = $c;
 					return $c;
@@ -759,17 +886,12 @@ sub locale {
 			}
 		}
 
-		eval {
-			require HTTP::BrowserDetect;
-
+		if(eval { require HTTP::BrowserDetect; } ) {
 			HTTP::BrowserDetect->import();
-		};
-
-		unless($@) {
 			my $browser = HTTP::BrowserDetect->new($agent);
 
 			if($browser && $browser->country()) {
-				my $c = Locale::Object::Country->new(code_alpha2 => $browser->country());
+				my $c = $self->_code2country($browser->country());
 				if($c) {
 					$self->{_locale} = $c;
 					return $c;
@@ -787,7 +909,7 @@ sub locale {
 		my $c;
 		eval {
 			local $SIG{__WARN__} = sub { die $_[0] };
-			$c = Locale::Object::Country->new(code_alpha2 => $country);
+			$c = $self->_code2country($country);
 		};
 		unless($@) {
 			if($c) {
@@ -800,7 +922,7 @@ sub locale {
 	# Try mod_geoip
 	if(defined($ENV{'GEOIP_COUNTRY_CODE'})) {
 		$country = $ENV{'GEOIP_COUNTRY_CODE'};
-		my $c = Locale::Object::Country->new(code_alpha2 => $country);
+		my $c = $self->_code2country($country);
 		if($c) {
 			$self->{_locale} = $c;
 			return $c;
@@ -809,6 +931,83 @@ sub locale {
 	return ();	# returns undef
 }
 
+# Wrapper to Locale::Language::code2language which makes use of the cache
+sub _code2language
+{
+	my ($self, $code) = @_;
+
+	return unless($code);
+	if($self->{_logger}) {
+		if(defined($self->{_country})) {
+			$self->{_logger}->trace("_code2language $code, country " . $self->{_country});
+		} else {
+			$self->{_logger}->trace("_code2language $code");
+		}
+	}
+	unless($self->{_cache}) {
+		return Locale::Language::code2language($code);
+	}
+	my $from_cache = $self->{_cache}->get("code2language/$code");
+	if($from_cache) {
+		if($self->{_logger}) {
+			$self->{_logger}->trace("_code2language found in cache $from_cache");
+		}
+		return $from_cache;
+	}
+	if($self->{_logger}) {
+		$self->{_logger}->trace('_code2language not in cache, storing');
+	}
+	return $self->{_cache}->set("code2language/$code", Locale::Language::code2language($code), '1 month');
+}
+
+# Wrapper to Locale::Object::Country allowing for persistance to be added
+sub _code2country
+{
+	my ($self, $code) = @_;
+
+	return unless($code);
+	if($self->{_logger}) {
+		if($self->{_country}) {
+			$self->{_logger}->trace("_code2country $code, country " . $self->{_country});
+		} else {
+			$self->{_logger}->trace("_code2country $code");
+		}
+	}
+	return Locale::Object::Country->new(code_alpha2 => $code);
+}
+
+# Wrapper to Locale::Object::Country->name which makes use of the cache
+sub _code2countryname
+{
+	my ($self, $code) = @_;
+
+	return unless($code);
+	if($self->{_logger}) {
+		$self->{_logger}->trace("_code2countryname $code");
+	}
+	unless($self->{_cache}) {
+		my $country = $self->_code2country($code);
+		if(defined($country)) {
+			return $country->name;
+		}
+		return;
+	}
+	my $from_cache = $self->{_cache}->get("code2countryname/$code");
+	if($from_cache) {
+		if($self->{_logger}) {
+			$self->{_logger}->trace("_code2countryname found in cache $from_cache");
+		}
+		return $from_cache;
+	}
+	if($self->{_logger}) {
+		$self->{_logger}->trace('_code2countryname not in cache, storing');
+	}
+	my $country = $self->_code2country($code);
+	if(defined($country)) {
+		return $self->{_cache}->set("code2countryname/$code", $country->name, '1 month');
+	}
+}
+
 =head1 AUTHOR
 
 Nigel Horne, C<< <njh at bandsman.co.uk> >>
@@ -863,7 +1062,7 @@ L<http://search.cpan.org/dist/CGI-Lingua/>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2010-2014 Nigel Horne.
+Copyright 2010-2015 Nigel Horne.
 
 This program is released under the following licence: GPL
 
